#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <SparkFunTMP102.h>
#include <Adafruit_MPR121.h>
#include <Adafruit_NeoPixel.h>

enum Pinout {
  DRIVE2 = 0,
  DRIVE1 = 1,
  LED_W = 2,
  WS2812_DISPLAY = 2,
  LED_Y = 3,
  EN_MEASURE_CURRENT = 3,
  LED_B = 4,
  LED_G = 5,
  LED_R = 6,
  MATRIX_IN1 = 7,
  SBU1 = 8,
  SBU2 = 9,
  VBUS_DET = 10,
  BOOT2 = 11,
  I2C_OLED_SDA = 12,
  I2C_OLED_SCL = 13,
  WS2811 = 14,
  TX_EN = 15,
  TX = 16,
  RX = 17,
  MATRIX_IN2 = 18,
  REED3 = 19,
  REED4 = 20,
  REED2 = 21,
  REED1 = 22,
  MATRIX_IN3 = 23,
  DIGOUT1 = 24,
  DIGOUT2 = 25,
  ANALOG_IN1 = 26,
  CURRENT2 = 27,
  CURRENT1 = 28,
  MEASURE_VCC = 29,
};

const int SCREEN_ADDRESS = 0x3C;
Adafruit_SSD1306 display(128, 64, &Wire, -1);
bool hasDisplay;

#include "logo.h"

TMP102 sensor0, sensor1;

Adafruit_MPR121 cap0, cap1;
bool capAvailable[2];

Adafruit_NeoPixel pixelsDisplay(60, WS2812_DISPLAY, NEO_GRB + NEO_KHZ800);

static bool initial = true;

void setup() {
  rp2040.enableDoubleResetBootloader();

  pinMode(DIGOUT1, OUTPUT);
  digitalWrite(DIGOUT1, LOW);

  Serial.begin(9600);  // USB serial, baudrate is ignored
  if (0) {
    while (!Serial)
      ;
    Serial.println("a1");
  }

  pinMode(DRIVE1, OUTPUT); digitalWrite(DRIVE1, LOW);
  pinMode(DRIVE2, OUTPUT); digitalWrite(DRIVE2, LOW);
  pinMode(WS2812_DISPLAY, OUTPUT);
  pinMode(3, OUTPUT);  digitalWrite(EN_MEASURE_CURRENT, LOW);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);

  pinMode(BOOT2, INPUT_PULLUP);
  pinMode(WS2811, INPUT_PULLUP);
  pinMode(REED1, OUTPUT);
  pinMode(REED2, OUTPUT);
  pinMode(REED3, OUTPUT);
  pinMode(REED4, OUTPUT);
  pinMode(MATRIX_IN1, INPUT_PULLUP);
  pinMode(MATRIX_IN3, INPUT_PULLUP);

  pinMode(TX_EN, OUTPUT);
  digitalWrite(TX_EN, LOW);

  pixelsDisplay.begin();
  pixelsDisplay.clear();
  pixelsDisplay.show();

  Serial1.setTX(TX);  // UART0
  Serial1.setRX(RX);
  Wire.setSDA(I2C_OLED_SDA);  // I2C0
  Wire.setSCL(I2C_OLED_SCL);

  Serial1.begin(115200);
  Wire.begin();
  Wire.setClock(100000);

  display.setRotation(2);
  hasDisplay = display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS);
  if (!hasDisplay) {
    Serial.println(F("SSD1306 allocation failed"));
  } else {
    display.clearDisplay();
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    display.display();
  }

  if (sensor0.begin(0x48)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor0.setConversionRate(2);
    sensor0.setExtendedMode(1);
    sensor0.wakeup();
  }
  if (sensor1.begin(0x49)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor1.setConversionRate(2);
    sensor1.setExtendedMode(1);
    sensor1.wakeup();
  }

  capAvailable[0] = cap0.begin(0x5a);
  capAvailable[1] = cap1.begin(0x5c);
  for (int i=0;i<2;i++) {
    if (!capAvailable[i])
      continue;
    auto& cap = (i == 0 ? cap0 : cap1);

    // disable it, again
    cap.writeRegister(MPR121_ECR, 0x0);

    // auto-config code (mostly) copied from Adafruit library
    cap.writeRegister(MPR121_AUTOCONFIG0, 0x0B);

    // correct values for Vdd = 3.3V
    cap.writeRegister(MPR121_UPLIMIT, 200);     // ((Vdd - 0.7)/Vdd) * 256
    cap.writeRegister(MPR121_TARGETLIMIT, 180); // UPLIMIT * 0.9
    cap.writeRegister(MPR121_LOWLIMIT, 130);    // UPLIMIT * 0.65

    // enable X electrodes and start MPR121
    // enable virtual proximity electrode, as well
    cap.writeRegister(MPR121_ECR, 0x8c | 0x30);
  }
}

void displayPrintHexFixed(uint32_t x, int width) {
  while (width > 1 && (x >> (4*(width-1))) == 0) {
    display.write('0');
    width--;
  }
  display.print(x, HEX);
}

void loop() {
  static uint8_t pwmValue = 32;
  static uint8_t pwmOn = false;
  bool changed = false;

  static int button_timer = 0;
  bool buttons[4];
  if (button_timer < 500)
    button_timer++;
  else {
    // Button Matrix:
    // 1: SDA, IN1
    // 2: SDA, IN3
    // 3: SCL, IN1
    // 4: SCL, IN4
    Wire.end();
    pinMode(I2C_OLED_SDA, OUTPUT);
    pinMode(I2C_OLED_SCL, INPUT_PULLUP);
    delayMicroseconds(1);
    buttons[0] = !digitalRead(MATRIX_IN1);
    buttons[1] = !digitalRead(MATRIX_IN3);
    pinMode(I2C_OLED_SDA, INPUT_PULLUP);
    pinMode(I2C_OLED_SCL, OUTPUT);
    delayMicroseconds(1);
    buttons[2] = !digitalRead(MATRIX_IN1);
    buttons[3] = !digitalRead(MATRIX_IN3);
    pinMode(I2C_OLED_SDA, INPUT_PULLUP);
    pinMode(I2C_OLED_SCL, INPUT_PULLUP);
    Wire.begin();

    if (buttons[0] && pwmValue > 0) {
      pwmValue--;
      changed = true;
    }
    if (buttons[1] && pwmValue < 255) {
      pwmValue++;
      changed = true;
    }
    if (buttons[2] && pwmOn) {
      pwmOn = false;
      changed = true;
    }
    if (buttons[3] && !pwmOn) {
      pwmOn = true;
      changed = true;
    }
  }

  if (changed || initial) {
    initial = false;

    analogWrite(DIGOUT1, pwmOn ? pwmValue : 0);

    int pixelsOn = pwmValue * 60 / 256;
    for (int i=0; i<pixelsDisplay.numPixels(); i++) {
      pixelsDisplay.setPixelColor(i,
        pixelsOn <= i ? pixelsDisplay.Color(0, 0, 0) :
        pwmOn ? pixelsDisplay.Color(0, 16, 0) :
        pixelsDisplay.Color(8, 8, 0));
    }
    pixelsDisplay.show();
  }


  static int led_timer = 0;
  static int led_state = 2;
  if (1)
    ;
  else if (led_timer < 500)
    led_timer++;
  else {
    led_timer = 0;
    led_state++;
    if (led_state == 8)
      led_state = 2;
    if (led_state > 2 && led_state-1 != EN_MEASURE_CURRENT && led_state-1 != WS2812_DISPLAY)
      digitalWrite(led_state-1, 0);
    if (led_state <= 6 && led_state != EN_MEASURE_CURRENT && led_state != WS2812_DISPLAY)
      digitalWrite(led_state, 1);
  }

  //if (rp2040.getCycleCount64() > 2 * 133*1000*1000 / 4 || changed) {
  static int display_timer = 0;
  if (display_timer < 500 && !changed)
    display_timer++;
  else {
    display_timer = 0;

    if (1) {
      display.fillRect(67, 0, 128-67, 64, SSD1306_BLACK);
      display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    }
    display.fillRect(0, 0, 67, 64, SSD1306_BLACK);

    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(0, 1);
    display.print(F("PWM: "));
    display.print(pwmValue);
    display.println(" / 256");

    if (0) {
      display.print(F("SW:   "));
      for (auto pressed : buttons)
        display.print(pressed ? 'X' : '_');
      display.println();
    }

    auto temp0 = sensor0.readTempC();
    auto temp1 = sensor1.readTempC();
    display.print(F("Temp: "));
    display.println(temp0, 2);
    display.print(F("      "));
    display.println(temp1, 2);

    if (0) {
      uint16_t touched0 = capAvailable[0] ? cap0.touched() : 0;
      uint16_t touched1 = capAvailable[1] ? cap1.touched() : 0;
      display.print(F("T:"));
      if (capAvailable[0])
        displayPrintHexFixed(touched0, 4);
      else
        display.write("----");
      if (capAvailable[1])
        displayPrintHexFixed(touched1, 4);
      else
        display.write("----");
      display.println();
    }

    display.setCursor(3, 57);
    display.println(F("---   +++ OFF     ON"));

    display.display();
  }

  if (!digitalRead(BOOT2)) {
    analogWrite(LED_R, 128);
  } else {
    analogWrite(LED_R, 0);
  }

  delay(1);
}